
flat varying vec4 labvalue;
flat varying vec2 midtex;
flat varying float translucent;
flat varying float emissive;
varying vec4 lmtexcoord;
varying vec4 color;
varying vec3 normalMat;

#if defined NORMAL_MAPS || defined POM
varying vec4 normalMat2;

varying vec4 tangent;
    #if defined POM
        #include "/lib/pom.glsl"
    #endif
#endif

// Constants for default values
const float DEFAULT_REFLECTANCE = 0.04;
const float SSS_MIN_THRESHOLD = 0.0;
const float SSS_MAX = 64.5 / 255.0;
const float EMISSIVE_MAX = 255.0;

#ifndef ENTITY

    #ifdef LAB_PBR
// Helper function for the LAB PBR path
vec4 getSpecularPropertiesLAB(vec2 tcoord, vec3 albedo, float emissive)
{
    vec4 specularTex = textureLod(specular, tcoord, 0.0);

    float smoothness = specularTex.x;
    float reflectance = specularTex.y > 0.0 ? specularTex.y : DEFAULT_REFLECTANCE;
    float sss = specularTex.z > 0.0
                    ? ((64.5 + ((specularTex.z * 255.0) * (255.0 - 64.5)) / 255.0) / 255.0)
                    : SSS_MIN_THRESHOLD;

    float emissives = (specularTex.w * EMISSIVE_MAX < EMISSIVE_MAX) ? specularTex.w : 0.0;
    if (emissives == 0.0)
    {
        emissives = emissive * pow(luma(albedo), 2.0);
    }

    return vec4(sss, emissives, reflectance, smoothness);
}
    #endif // LAB_PBR

    #ifndef LAB_PBR
// Helper function for the non-LAB PBR path
vec4 getSpecularPropertiesNonLAB(vec2 tcoord, vec2 midtex, vec3 albedo, float emissive)
{
    float sss = labvalue.z;
    float emissives = emissive * pow(luma(albedo), 2.0);
        #ifdef ENTITY
    if (entityId == 1803)
    {
        sss = 1.0;
    }

    if (entityId == 1602)
    {
        emissives = 1.0;
    }
        #endif
    // Noise-based sampling to refine albedo

    vec3 albedo2;
        #ifdef LOD_FIX
    vec2 deltaPos = (1.0 / textureSize(texture, 0)) * noise_standard(gl_FragCoord.xy) * 4.0;

    vec2 offsets[6] = vec2[](
        vec2(deltaPos.x, 0.0),
        vec2(-deltaPos.x, 0.0),
        vec2(0.0, deltaPos.y),
        vec2(0.0, -deltaPos.y),
        vec2(deltaPos),
        vec2(-deltaPos));

    vec3 albedoAccum = texture(texture, midtex).rgb;
    for (int i = 0; i < 4; ++i)
    {
        albedoAccum += texture(texture, midtex + offsets[i]).rgb;
    }
    albedoAccum += texture(texture, midtex + offsets[4]).rgb;
    albedoAccum += texture(texture, midtex + offsets[5]).rgb;
    albedo2 = albedoAccum / 6.0;
        #else
    albedo2 = textureLod(texture, tcoord, 2.0).rgb;
        #endif

    float lumaValue = luma(albedo);
    float lumaLod = luma(albedo2);
    float modifier = clamp(pow(abs(lumaValue - lumaLod), 0.25) +
                               pow(abs(lumaValue - lumaLod), 0.5),
                           0.0, 1.0);

    float reflectance = mix(labvalue.x * 1.2, labvalue.x * 0.7, modifier);
    float smoothness = mix(labvalue.y * 1.2, labvalue.y * 0.7, modifier);

    vec3 sssParams = mix(vec3(16.0, 47.0, labvalue.w), vec3(64.5, 190.5, sss), step(0.0, sss));
    sss = (sssParams.x + sssParams.y * mix(sssParams.z * 1.2, sssParams.z, modifier)) / 255.0;

    // Final adjustment for high reflectance values
    float condition = step(0.9, labvalue.x);
    reflectance = mix(reflectance, labvalue.x, condition);
    smoothness = mix(smoothness, clamp(mix(labvalue.y * 1.2, labvalue.y * 0.7, modifier), 0.0, 1.0), condition);

    return vec4(sss, emissives, reflectance, smoothness);
}
    #endif // not LAB_PBR

#endif // not ENTITY

// Main function to obtain specular properties.
vec4 getSpecularProperties(vec2 tcoord, vec2 midtex, vec3 albedo, float translucent, float emissive)
{
    // Sample base albedo
    albedo = textureLod(texture, tcoord, 0.0).rgb;

#ifndef ENTITY
    #ifdef LAB_PBR
    return getSpecularPropertiesLAB(tcoord, albedo, emissive);
    #else
    return getSpecularPropertiesNonLAB(tcoord, midtex, albedo, emissive);
    #endif
#else
    // Fallback default values if ENTITY is defined
    return vec4(0.0, 0.0, 0.02, 0.1);
#endif
}

#ifdef GENERATED_NORMALS
    #include "/lib/normal_generator.glsl"
#endif

vec3 calculateNormal(vec2 tcoord, vec2 lmtexcoord, mat3 tbnMatrix, vec2 midtex, inout float ao)
{
    vec3 normalTex;
    vec3 normalout;

#ifdef GENERATED_NORMALS
    normalTex = normalGenerator(lmtexcoord.xy, midtex);
    normalout.xy = normalTex.xy * 2.0 - 0.996078431372549;
    normalout.z = sqrt(clamp(1.0 - dot(normalout.xy, normalout.xy), 0.0, 1.0));

#else
    normalTex = texture(normals, tcoord, 0).rgb;

    ao *= normalTex.b;

    normalout.xy = normalTex.xy * 2.0 - 0.996078431372549;
    normalout.z = sqrt(clamp(1.0 - dot(normalout.xy, normalout.xy), 0.0, 1.0));
#endif

    float bumpmultiplier = BUMP_MULT;
    float bumpmult = (1.0 - wetness * 0.95) * bumpmultiplier;

    normalout = normalout * vec3(bumpmult, bumpmult, bumpmult) + vec3(0.0f, 0.0f, 1.0f - bumpmult);

    normalout = normalize(normalout * tbnMatrix);

    return normalout;
}

#if defined POM
float interleaved_gradientNoise()
{
    return fract(52.9829189 * fract(0.06711056 * gl_FragCoord.x + 0.00583715 * gl_FragCoord.y) + frameTimeCounter * 51.9521);
}

vec2 applyPOM(vec2 tcoord, vec3 fragpos, inout vec3 pomnormal)
{
    // Compute TBN matrix
    vec3 tangent2 = normalize(cross(tangent.rgb, normalMat2.rgb) * tangent.w);
    mat3 tbnMatrix2 = mat3(
        tangent.x, tangent2.x, normalMat2.x,
        tangent.y, tangent2.y, normalMat2.y,
        tangent.z, tangent2.z, normalMat2.z);

    // Apply POM
    vec2 newTcoord = pomColorAndNormal(tcoord, fragpos, interleaved_gradientNoise(), tbnMatrix2, pomnormal);

    return newTcoord;
}
#endif
vec4 getMaterialProperties(vec2 tcoord, vec3 albedoRGB)
{
    return getSpecularProperties(tcoord, midtex, albedoRGB, translucent, emissive);
}
vec3 calculateNormalWrapper(vec2 tcoord, float ao, vec3 fragpos, vec3 normalMat2)
{

    vec3 normal;
#ifdef NORMAL_MAPS

    vec3 tangent2 = normalize(cross(tangent.rgb, normalMat2) * tangent.w);
    mat3 tbnMatrix2 = mat3(
        tangent.x, tangent2.x, normalMat2.x,
        tangent.y, tangent2.y, normalMat2.y,
        tangent.z, tangent2.z, normalMat2.z);

    normal = viewToWorld(calculateNormal(tcoord, lmtexcoord.xy, tbnMatrix2, midtex, ao));
#else
    normal = normalMat;
#endif

#ifdef NORMAL_FIX
    vec3 cameraVector = normalize(fragpos);
    vec3 vnormal = worldToView(normalMat);
    normal = viewToWorld(faceforward(worldToView(normal), cameraVector, vnormal));
#endif

    return normal;
}
vec2 computeAdvancedLighting(vec2 lightmap, vec3 fragpos, vec3 normal)
{
#ifdef DLM
    // Compute derivatives
    vec3 dx = dFdx(fragpos);
    vec3 dy = dFdy(fragpos);
    vec3 dz = cross(dx, dy) * 0.01;

    vec2 derivation = 256.0 * vec2(dFdx(lightmap.x), dFdy(lightmap.x));
    vec3 direction = normalize(derivation.x * dx + derivation.y * dy + dz);

    float ndotd = dot(worldToView(normal), direction);
    float power = 1.0 - ndotd * lightmap.x;
    float derivationLength = min(length(derivation) * 10.0, 1.0);

    lightmap.x = mix(lightmap.x, pow(lightmap.x, power), derivationLength);
#endif

    return lightmap;
}

void outputToFramebuffer(vec4 albedo, vec3 normal, vec2 lightmap, vec4 materialProps, float ao)
{

    gl_FragData[0] = albedo; // Albedo
    gl_FragData[1].rgba = vec4(
        encodeNormal(normal), encodeVec2(lightmap), encodeVec2(vec2(ao, materialProps.x))); // Normals and materials
    gl_FragData[2].r = encodeVec2(materialProps.zw);
#ifdef ENTITY
    #ifdef NO_ENTITY_TAA
    gl_FragData[3].g = 1.0;
    #endif
#endif
}
#ifdef ENTITY
    #ifdef NO_ENTITY_TAA
    /* RENDERTARGETS: 0,2,3,5 */
    #else
    /* RENDERTARGETS: 0,2,3 */
    #endif

#else
/* RENDERTARGETS: 0,2,3 */
#endif
void main()
{
    vec2 tcoord = lmtexcoord.xy;
    vec2 lightmap = lmtexcoord.zw;

    // Compute fragment position in screen space
    vec3 fragpos = toScreenSpace(gl_FragCoord.xyz * vec3(texelSize, 1.0));
    vec3 pomnormal = vec3(0.0);

#ifdef NORMAL_MAPS
    pomnormal = normalMat2.rgb;
#endif
// Parallax Occlusion Mapping (POM)
#if defined POM
    tcoord = applyPOM(tcoord, fragpos, pomnormal);

#endif

    // Fetch textures and material properties
    vec4 albedo = texture(texture, tcoord, 0);

    vec4 materialProps = getMaterialProperties(tcoord, albedo.rgb);

    // Apply color modifications
    albedo.rgb *= color.rgb;
#ifdef ENTITY
    //   if (entityId == 1600) // not a good fix, will sort this later.tm
    //   {
    //     discard;
    //   }

    albedo.rgb = mix(albedo.rgb, entityColor.rgb, entityColor.a);

#endif

#if defined POM
    vec3 normal = (viewToWorld(pomnormal));
#else
    vec3 normal = calculateNormalWrapper(tcoord, color.a, fragpos, pomnormal.rgb);
#endif

// Additional lighting computations
#ifdef PBR
    lightmap = computeAdvancedLighting(lightmap, fragpos, normal);
    lightmap.x = min(lightmap.x + materialProps.y, 1.0);
#endif
#ifdef ENTITY
    #ifdef NAMETAG_FIX
    if (entityId == 1500)
    {
        normal = viewToWorld(vec3(0, 0, 1));

        albedo.rgb *= 2.0;
        float mask1 = step(0.0, lmtexcoord.r) * (1.0 - step(0.02, lmtexcoord.r)) * step(0.0351, lmtexcoord.g) * (1.0 - step(0.0666, lmtexcoord.g));
        float mask2 = step(0.49, lmtexcoord.r) * (1.0 - step(0.51, lmtexcoord.r)) * step(0.49, lmtexcoord.g) * (1.0 - step(0.51, lmtexcoord.g));

        albedo.a *= 1.0 - max(mask1, mask2);
    }
    #endif
#endif
    albedo.rgb = toLinear(albedo.rgb);
    // Output to framebuffer
    outputToFramebuffer(albedo, normal, lightmap, materialProps, color.a);
}
