

flat varying vec3 ambientUp;
flat varying vec3 ambientDown;
flat varying vec3 ambientLeft;
flat varying vec3 ambientRight;
flat varying vec3 ambientB;
flat varying vec3 ambientF;
flat varying float ambientA;
flat varying vec4 sunColor;
varying vec2 texcoord;

#include "/lib/sky_function.glsl"






vec3 GenerateUnitVector(vec2 xy)
{
    xy.x *= radians(360.0);
    xy.y = xy.y * 2.0 - 1.0;
    float s = sin(xy.x);
    float c = cos(xy.x);
    return vec3(vec2(s, c) * sqrt(1.0 - xy.y * xy.y), xy.y);
}

void calculateLighting2(
    out vec3 ambientUp,
    out vec3 ambientLeft,
    out vec3 ambientRight,
    out vec3 ambientB,
    out vec3 ambientF,
    out vec3 ambientDown,
    out float ambientA,
    out vec4 sunColor)
{
    // Precompute sun visibility: clamp once, then scale and square.
    float clampedSunElevation = clamp(sunElevation, 0.0, 0.05);
    float t = clampedSunElevation / 0.05;
    float sunVis = t * t;
    sunColor = vec4(sunLight, sunVis);

    // Define sample counts and precompute their inverse.
    const ivec2 samples = ivec2(32, 16);
    const vec2 invSamples = 1.0 / vec2(samples);
    const float sampleWeight = 2.0 / float(samples.x * samples.y);

    // Initialize ambient accumulators.
    ambientRight = ambientUp = ambientF = vec3(0.0);
    ambientLeft = ambientDown = ambientB = vec3(0.0);

    // Loop over the sample grid.
    for (int x = 0; x < samples.x; ++x)
    {
        for (int y = 0; y < samples.y; ++y)
        {
            // Compute UV coordinate once per sample.
            vec2 uv = (vec2(x, y) + 0.5) * invSamples;
            vec3 dir = GenerateUnitVector(uv);

            // Ensure upward direction by taking the absolute value of Y.
            dir.y = abs(dir.y);

            // Sample the sky texture.
            vec3 skySample = skyFromTex2(dir, colortex6).rgb;

            // Since dir components are in [-1,1], max(x, 0.0) is equivalent to clamp(x, 0.0, 1.0)
            ambientRight += skySample * max(dir.x, 0.0);
            ambientUp += skySample * max(dir.y, 0.0);
            ambientF += skySample * max(dir.z, 0.0);

            ambientLeft += skySample * max(-dir.x, 0.0);
            ambientB += skySample * max(-dir.z, 0.0);
        }
    }

    // Apply the constant weight to the ambient contributions.
    ambientRight *= sampleWeight;
    ambientUp *= sampleWeight;
    ambientF *= sampleWeight;
    ambientLeft *= sampleWeight;
    ambientB *= sampleWeight;

    // Super simple fake skylight bounce.
    const float fakeBounceAlbedo = 0.5;
    ambientRight += ambientUp * fakeBounceAlbedo * 0.5;
    ambientF += ambientUp * fakeBounceAlbedo * 0.5;
    ambientLeft += ambientUp * fakeBounceAlbedo * 0.5;
    ambientDown += ambientUp * fakeBounceAlbedo;
    ambientB += ambientUp * fakeBounceAlbedo * 0.5;

    // Compute lightDir: if sunVis is effectively nonzero, use 1.0; otherwise, -1.0.
    float lightDir = (sunVis >= 1e-5) ? 1.0 : -1.0;

#ifndef OVERWORLD
    // Reassign sunLight in non-OVERWORLD mode.
    vec3 sunLightLocal = ambientUp * 512.0;
#else
    vec3 sunLightLocal = sunLight;
#endif

    // Fake bounced sunlight.
    vec3 bouncedSun = sunLightLocal * 0.001;
    ambientUp += bouncedSun * clamp(-lightDir * sunVec.y + 3.0, 0.0, 4.0);
    ambientLeft += bouncedSun * clamp(lightDir * sunVec.x + 3.0, 0.0, 4.0);
    ambientRight += bouncedSun * clamp(-lightDir * sunVec.x + 3.0, 0.0, 4.0);
    ambientB += bouncedSun * clamp(-lightDir * sunVec.z + 3.0, 0.0, 4.0);
    ambientF += bouncedSun * clamp(lightDir * sunVec.z + 3.0, 0.0, 4.0);
    ambientDown += bouncedSun * clamp(lightDir * sunVec.y + 3.0, 0.0, 4.0) * 0.7;

    // Compute ambient overall luminance.
    ambientA = luma((ambientUp + ambientLeft + ambientRight + ambientB + ambientF + ambientDown) / 6.0);
}

void main()
{

    texcoord = gl_MultiTexCoord0.xy;

    calculateLighting2(ambientUp, ambientLeft, ambientRight, ambientB, ambientF, ambientDown, ambientA, sunColor);

    gl_Position = ftransform();
}
